"""Hyperparameter checker for pytorch checks whether important hyperparameters are set."""
import astroid
from pylint.interfaces import IAstroidChecker
from pylint.lint import PyLinter
from dslinter.checkers.hyperparameters import HyperparameterChecker
from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.hyperparameters_helper import check_module_with_library


class HyperparameterTensorflowChecker(HyperparameterChecker):
    """Hyperparameter checker for pytorch checks whether important hyperparameters are set."""

    __implements__ = IAstroidChecker

    name = "hyperparameters-tensorflow"
    priority = -1
    msgs = {
        "R5507": (
            "Some of the important hyperparameters(learning rate, batch size, momentum, and weight decay) is not set in the program.",
            "hyperparameters-tensorflow",
            "Important hyperparameters should be set in the program."
        )
    }

    options = (
        (
            "strict_hyperparameters_tensorflow",
            {
                "default": False,
                "type": "yn",
                "metavar": "<y_or_n>",
                "help": "Force that all parameters of learning algorithms are set.",
            },
        ),
    )

    def __init__(self, linter: PyLinter = HyperparameterChecker):
        super().__init__(linter)
        self.HYPERPARAMETER_RESOURCE = "hyperparameters_tensorflow_dict.pickle"
        self.MESSAGE = "hyperparameters-tensorflow"
        self.HYPERPARAMETERS_MAIN = {
            # training
            # optimizer
            "Adadelta": {"positional": 5,"keywords": ["learning_rate"]},
            "Adagrad": {'positional': 5,"keywords": ["learning_rate"]},
            "Adam": {'positional': 7, "keywords": ["learning_rate"]},
            "Adamax": {'positional': 6, "keywords": ["learning_rate"]},
            "Ftrl": {'positional': 9, "keywords": ["learning_rate"]},
            "Nadam": {'positional': 6, "keywords": ["learning_rate"]},
            "RMSprop": {'positional': 7, "keywords": ["learning_rate", "momentum"]},
            "SGD": {'positional': 5, "keywords": ["learning_rate", "momentum"]},
        }
        # pylint: disable = invalid-name
        self.HYPERPARAMETERS_MAIN_2 = {
            "fit": {"positional": 19,"keywords": ["batch_size"]},
        }
        self.LIBRARY = "tensorflow"
        self.hyperparams_all_in_function = {
            "fit": {
                "positional": 19,
                "keywords": ["x", "y", "batch_size", "epochs", "verbose",
                                "callbacks", "validation_split", "validation_data", "shuffle",
                                "class_weight", "sample_weight", "initial_epoch", "steps_per_epoch",
                                "validation_steps", "validation_batch_size", "validation_freq",
                                "max_queue_size", "workers", "use_multiprocessing"]
            },
        }

    def visit_call(self, node: astroid.Call):
        """
        When a Call node is visited, check whether hyperparameters are set.

        In strict mode, all hyperparameters should be set.
        In non-strict mode, function calls to learning functions should either contain all
        hyperparameters defined in HYPERPARAMETERS_MAIN or have at least one hyperparameter defined.

        :param node: Node which is visited.
        """
        try:
            if(
                hasattr(node, "func")
                and hasattr(node.func, "name")
            ):
                self.hyperparameter_in_class(node, node.func.name)
            if(
                check_module_with_library(node, "tensorflow")
                and hasattr(node, "func")
                and hasattr(node.func, "attrname")
                and node.func.attrname == "fit"
            ):
                self.hyperparameters_in_function(node)

        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    def hyperparameters_in_function(self, node: astroid.Call):
        """Check whether we have required hyperparamter in a specific function."""
        function_name = node.func.attrname

        if function_name in self.hyperparams_all_in_function:  # pylint: disable=unsupported-membership-test
            if self.config.strict_hyperparameters_tensorflow: # strict mode
                # pylint: disable = line-too-long
                if not self.has_required_hyperparameters(node, self.hyperparams_all_in_function, function_name):
                    self.add_message(self.MESSAGE, node=node)
            else:  # non-strict mode
                if (
                    function_name in self.HYPERPARAMETERS_MAIN_2
                    and not self.has_required_hyperparameters(node, self.HYPERPARAMETERS_MAIN_2, function_name)
                ):
                    self.add_message(self.MESSAGE, node=node)
                elif len(node.args) == 0 and node.keywords is None:
                    self.add_message(self.MESSAGE, node=node)



"""Utility module for handling exceptions thrown in checkers."""
import astroid
from pylint.checkers import BaseChecker

from dslinter.utils.ast import ASTUtil


class ExceptionHandler:
    """Utility class for handling exceptions thrown in checkers."""

    # pylint: disable=too-few-public-methods

    @staticmethod
    def handle(checker: BaseChecker, node: astroid.node_classes.NodeNG):
        """
        Handle a generic exception thrown in a checker by printing an error message.

        :param checker: Checker where the exception is thrown from.
        :param node: Node which is visited while the exception is thrown.
        """
        module = ASTUtil.search_module(node)
        print(
            "ERROR: Could not finish processing the checker {} on module {} at line {}. Continuing.".format(
                checker.name, module.name, node.lineno
            )
        )



"""Class which tests the TypeInference utils class."""
import astroid

from dslinter.utils.type_inference import TypeInference


class TestTypeInference:
    """Class which tests the TypeInference utils class."""

    # def test_infer_types(self):
    #     """Test the infer_types method."""
    #     code = "a = 'b'; a.join([])"
    #     module_node = astroid.parse(code)
    #     node_type = astroid.Call
    #     result = TypeInference.infer_types(module_node, node_type, lambda x: x.func.expr.name)
    #
    #     assert result == {module_node.body[1].value: "builtins.str"}

    def test_add_reveal_type_calls(self):
        """Test the add_reveal_type_calls() method with a single expression."""
        code = "a = b.c(d)"
        nodes = [astroid.parse(code).body[0].value]

        result = TypeInference.add_reveal_type_calls(code, nodes, lambda node: node.func.expr.name)
        assert result == "a = b.c(d); reveal_type(b)"

    def test_add_reveal_type_calls_multiple(self):
        """Test the add_reveal_type_calls() method with multiple expressions."""
        code = "a = b.c(d)\nx = 5 \ne = f.g()"
        tree = astroid.parse(code)
        nodes = [tree.body[0].value, tree.body[2].value]

        result = TypeInference.add_reveal_type_calls(code, nodes, lambda node: node.func.expr.name)
        assert result == "a = b.c(d); reveal_type(b)\nx = 5 \ne = f.g(); reveal_type(f)"

    def test_run_mypy_success(self):
        """Test if mypy is ran successfully on some correct code."""
        result = TypeInference.run_mypy("a = 5")
        assert result == "Success: no issues found in 1 source file\n"

    def test_run_mypy_error(self):
        """Test if mypy returns an error when code is incorrect."""
        result = TypeInference.run_mypy("a: str = 5")
        assert result.splitlines()[1] == "Found 1 error in 1 file (checked 1 source file)"

    def test_parse_mypy_result(self):
        """Test if the parse_mypy_result method returns the correct type."""
        mypy_result = "<string>:1: note: Revealed type is 'builtins.int'"
        assert TypeInference.parse_mypy_result(mypy_result) == [(1, "'builtins.int'")]

    def test_parse_mypy_result_multiple(self):
        """Test if the parse_mypy_result method returns the correct types."""
        mypy_result = "<string>:1: note: Revealed type is 'builtins.int'"
        mypy_result += "\n<string>:2: note: Revealed type is 'builtins.str'"

        result = TypeInference.parse_mypy_result(mypy_result)
        assert result == [(1, "'builtins.int'"), (2, "'builtins.str'")]

    def test_combine_nodes_with_inferred_types(self):
        """Test if combine_nodes_with_inferred_types returns a correct dict on a single line."""
        nodes = [astroid.extract_node("a.b()")]
        types = [(1, "builtins.str")]
        result = TypeInference.combine_nodes_with_inferred_types(nodes, types)
        assert result == {nodes[0]: types[0][1]}

    def test_add_reveal_type_calls_block(self):
        """Test if the reveal_type() call is added to the body of a block statement."""
        code = "y = ''\nfor x in y.join([]):\n\tpass"
        nodes = [astroid.parse(code).body[1].iter]

        result = TypeInference.add_reveal_type_calls(code, nodes, lambda node: node.func.expr.name)
        assert result == "y = ''\nfor x in y.join([]):\n\tpass; reveal_type(y)"

    # def test_type_inference_in_for_loop_definition(self):
    #     """Test if type inference works for Calls inside for loop definitions."""
    #     code = "y = ''\nfor x in y.join([]):\n\tpass"
    #     module_node = astroid.parse(code)
    #     node_type = astroid.Call
    #     result = TypeInference.infer_types(module_node, node_type, lambda x: x.func.expr.name)
    #
    #     assert result == {module_node.body[1].iter: "builtins.str"}



import astroid


def check_main_module(module: astroid.Module) -> bool:
    for node in module.body:
        if isinstance(node, astroid.nodes.If) and hasattr(node, "test"):
            if_compare_node = node.test
            if(
                hasattr(if_compare_node, "left")
                and hasattr(if_compare_node.left, "name")
                and if_compare_node.left.name == "__name__"
                and hasattr(if_compare_node, "ops")
                and len(if_compare_node.ops) > 0
                and hasattr(if_compare_node.ops[0][1], "value")
                and if_compare_node.ops[0][1].value == '__main__'
            ):
                return True
    return False


def has_import(node: astroid.Import, library_name: str):
    for name, _ in node.names:
        if name == library_name:
            return True
    return False


def has_importfrom_sklearn(node: astroid.ImportFrom):
    if node.modname[:7] == "sklearn":
        return True
    return False




"""Checker which checks optimizer.zero_grad() is used when loss_fn.backward() and optimizer.step() are used."""
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.utils.exception_handler import ExceptionHandler


class GradientClearPytorchChecker(BaseChecker):
    """Checker which checks optimizer.zero_grad() is used when loss_fn.backward() and optimizer.step() are used."""

    __implements__ = IAstroidChecker

    name = "gradient-clear-pytorch"
    priority = -1
    msgs = {
        "W5517": (
            "The optimizer.zero_grad() is not used in pytorch code when loss_fn.backward() and optimizer.step() are used.",
            "gradient-clear-pytorch",
            "The loss_fn.backward() and optimizer.step() should be used together with optimizer.zero_grad() to clear gradients."
        )
    }
    options = ()

    def visit_for(self, for_node: astroid.For):
        """
        When a For node is visited, check whether it violated the rule in this checker.
        :param for_node: The node which is visited.
        """
        try:
            _has_zero_grad = False
            _has_backward = False
            _has_step = False
            for n in for_node.body:
                if(
                    isinstance(n, astroid.Expr)
                    and hasattr(n.value, "func")
                    and hasattr(n.value.func, "attrname")
                ):
                    if n.value.func.attrname == "zero_grad":
                        _has_zero_grad = True
                    if n.value.func.attrname == "backward":
                        _has_backward = True
                    if n.value.func.attrname == "step":
                        _has_step = True
            if _has_backward is True and _has_step is True and _has_zero_grad is False:
                self.add_message("gradient-clear-pytorch", node=for_node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, for_node)




"""Utility module for reading resources."""

import pickle
from typing import Dict, List, Union

from pkg_resources import resource_stream


class Resources:
    """Utility class for reading resources."""

    __RESOURCES_PACKAGE = "dslinter.resources"

    @staticmethod
    def get_hyperparameters(hyperparameter_resource) -> Dict[str, Dict[str, Union[int, List[str]]]]:
        """
        Get the hyperparameters resource.

        :return: Dict with every learning algorithm from scikit-learn as keys. Each value is a Dict
        containing the keys 'positional' and 'keywords' containing its amount of keywords and a list
        with the names of its keywords respectively.
        """
        return Resources.read_pickle(hyperparameter_resource)

    @staticmethod
    def read_pickle(file):
        """
        Read a pickled object from disk.

        :param file: File name of the pickle in the resource package.
        :return: Deserialized object.
        """
        return pickle.load(resource_stream(Resources.__RESOURCES_PACKAGE, file))



"""Hyperparameter checker for pytorch checks whether important hyperparameters are set."""
from pylint.interfaces import IAstroidChecker
from pylint.lint import PyLinter
from dslinter.checkers.hyperparameters import HyperparameterChecker


class HyperparameterPyTorchChecker(HyperparameterChecker):
    """Hyperparameter checker for pytorch checks whether important hyperparameters are set."""

    __implements__ = IAstroidChecker

    name = "hyperparameters-pytorch"
    priority = -1
    msgs = {
        "R5508": (
            "Some of the important hyperparameters(learning rate, batch size, momentum, and weight decay) is not set in the program.",
            "hyperparameters-pytorch",
            "Important hyperparameters should be set in the program."
        )
    }

    options = (
        (
            "strict_hyperparameters_pytorch",
            {
                "default": False,
                "type": "yn",
                "metavar": "<y_or_n>",
                "help": "Force that all parameters of learning algorithms are set.",
            },
        ),
    )

    def __init__(self, linter: PyLinter = HyperparameterChecker):
        super().__init__(linter)
        self.HYPERPARAMETER_RESOURCE = "hyperparameters_pytorch_dict.pickle"
        self.MESSAGE = "hyperparameters-pytorch"
        self.HYPERPARAMETERS_MAIN = {
            # dataloader
            "DataLoader": {"positional": 2,"keywords":["batch_size"]},
            # optimizer
            "Adadelta": {"positional": 5, "keywords": ["lr", "weight_decay"]},
            "Adagrad": {"positional": 6, "keywords": ["lr", "weight_decay"]},
            "Adam": {"positional": 6, "keywords": ["lr", "weight_decay"]},
            "AdamW": {"positional": 6, "keywords": ["lr", "weight_decay"]},
            "SparseAdam": {"positional": 4, "keywords": ["lr"]},
            "Adamax": {"positional": 5, "keywords": ["lr", "weight_decay"]},
            "ASGD": {"positional": 6, "keywords": ["lr", "weight_decay"]},
            "LBFGS": {"positional": 8, "keywords": ["lr"]},
            "NAdam": {"positional": 6, "keywords": ["lr", "weight_decay", "momentum_decay"]},
            "RAdam": {"positional": 5, "keywords": ["lr", "weight_decay"]},
            "RMSprop": {"positional": 7, "keywords": ["lr", "weight_decay", "momentum"]},
            "Rprop": {"positional": 4, "keywords": ["lr"]},
            "SGD": {"positional": 6, "keywords": ["lr", "weight_decay", "momentum"]},
        }
        self.LIBRARY = "pytorch"



"""Class which test ChainIndexingPandasChecker."""
import astroid
import pylint.testutils

import dslinter


class TestChainIndexingPandasChecker(pylint.testutils.CheckerTestCase):
    """Class which test ChainIndexingPandasChecker."""

    CHECKER_CLASS = dslinter.plugin.ChainIndexingPandasChecker

    def test_chain_indexing_on_dataframe1(self):
        """Message should be added if there is a chain indexing on pandas dataframe."""
        script = """
        import pandas as pd
        df = pd.DataFrame([[1,2,3],[4,5,6]])
        col = 1
        x = 0
        df[col][x] = 42 #@
        """
        module = astroid.parse(script)
        assign_node = astroid.extract_node(script)
        subscript_node = assign_node.targets[0]
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="chain-indexing-pandas", node=subscript_node)):
            self.checker.visit_module(module)
            self.checker.visit_subscript(subscript_node)

    def test_chain_indexing_on_dataframe2(self):
        """Message should be added if there is a chain indexing on pandas dataframe."""
        script = """
        import pandas as pd
        df = pd.DataFrame([[[1, 1],[2, 2],[3, 3]],[[4, 4],[5, 5],[6, 6]]])
        col = 1
        x = 0
        y = 0
        df[col][x][y] = 42 #@
        """
        module = astroid.parse(script)
        assign_node = astroid.extract_node(script)
        subscript_node = assign_node.targets[0]
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="chain-indexing-pandas", node=subscript_node)):
            self.checker.visit_module(module)
            self.checker.visit_subscript(subscript_node)

    def test_chain_indexing_on_normal_array(self):
        """No message should be added if there is a chain indexing on normal array."""
        script = """
        a = [[0, 0][0, 0]]
        a[0][1] = 2 #@
        """
        module = astroid.parse(script)
        assign_node = astroid.extract_node(script)
        subscript_node = assign_node.targets[0]
        with self.assertNoMessages():
            self.checker.visit_module(module)
            self.checker.visit_subscript(subscript_node)

    def test_indexing_on_dataframe(self):
        """No message should be added if there is a normal indexing on pandas dataframe."""
        script = """
        import pandas as pd
        df = pd.DataFrame([[1,2,3],[4,5,6]])
        col = 1
        x = 0
        df[col] = 42 #@        
        """
        module = astroid.parse(script)
        assign_node = astroid.extract_node(script)
        subscript_node = assign_node.targets[0]
        with self.assertNoMessages():
            self.checker.visit_module(module)
            self.checker.visit_subscript(subscript_node)

    def test_no_chain_indexing_on_dataframe(self):
        """No message should be added when there is no chain indexing on dataframe."""
        script = """
        import pandas as pd
        df = pd.DataFrame([[1,2,3],[4,5,6]])
        col = 1
        x = 0
        df.loc[x, col] = 42  #@     
        """
        module = astroid.parse(script)
        assign_node = astroid.extract_node(script)
        subscript_node = assign_node.targets[0]
        with self.assertNoMessages():
            self.checker.visit_module(module)
            self.checker.visit_subscript(subscript_node)




"""Checker which checks column is selected after the dataframe is imported."""
from collections import defaultdict

import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.utils.exception_handler import ExceptionHandler


class ColumnSelectionPandasChecker(BaseChecker):
    """Checker which checks column is selected after the dataframe is imported."""

    __implements__ = IAstroidChecker

    name = "column-selection-pandas"
    priority = -1
    msgs = {
        "R5504": (
            "There is no column selection after the dataframe is imported.",
            "column-selection-pandas",
            "Column should be selected after the dataframe is imported for better elaborating what to be expected in the downstream."
        )
    }
    options = ()

    _data_import_functions = ["read_csv", "read_table", "read_fwf", "read_excel", "read_xml"]

    def visit_module(self, module: astroid.Module):
        """
        Visit module and see whether the rule is violated.
        :param module:
        :return: None
        """
        try:
            self._traverse_content(module)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)

    def visit_functiondef(self, functiondef_node: astroid.FunctionDef):
        """
        Visit functiondef node and see whether the rule is violated.
        :param function_def_node:
        :return: None
        """
        try:
            self._traverse_content(functiondef_node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, functiondef_node)

    def _traverse_content(self, block):
        """Traverse through the block and see whether the rule is violated"""
        _dataframe_selected_dict = defaultdict(bool)
        _name_node_dict = defaultdict()
        for _, node in enumerate(block.body):
            if isinstance(node, astroid.Assign):
                self._process_assign_node(node, _dataframe_selected_dict, _name_node_dict)

        # check which dataframe is imported but not selected by columns
        for k, v in _dataframe_selected_dict.items():
            if v is False:
                self.add_message("column-selection-pandas", node=_name_node_dict[k])

    def _process_assign_node(self, node: astroid.Assign, _dataframe_selected_dict, _name_node_dict):
        """Check assign node to see if there is a imported dataframe or whether the datdframe columns are selected."""
        # check if there is a imported dataframe
        if(
            hasattr(node.value, "func")
            and hasattr(node.value.func, "attrname")
            and node.value.func.attrname in self._data_import_functions
            and hasattr(node.value.func, "expr")
            and hasattr(node.value.func.expr, "name")
            and node.value.func.expr.name in ["pandas", "pd"]
            and len(node.targets) > 0
            and hasattr(node.targets[0], "name")
        ):
            _imported_dataframe_name = node.targets[0].name
            _dataframe_selected_dict[_imported_dataframe_name] = False
            _name_node_dict[_imported_dataframe_name] = node
        # check if dataframe columns are selected
        if(
            hasattr(node, "value")
            and isinstance(node.value, astroid.Subscript)
            and hasattr(node.value, "value")
            and hasattr(node.value.value, "name")
            and node.value.value.name in _dataframe_selected_dict
            and hasattr(node.value, "slice")
            and isinstance(node.value.slice, astroid.nodes.List)
        ):
            _selected_dataframe_name = node.value.value.name
            _dataframe_selected_dict[_selected_dataframe_name] = True



"""Checker which checks whether there are possible invalid value unmasked."""
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.type_inference import TypeInference


class MaskMissingTensorflowChecker(BaseChecker):
    """Checker which checks whether there are possible invalid value unmasked."""

    __implements__ = IAstroidChecker

    name = "missing-mask-tensorflow"
    priority = -1
    msgs = {
        "W5513": (
            "The variable in tf.log() isn't wrapped with tf.clip_by_value().",
            "missing-mask-tensorflow",
            "Add a mask for possible invalid values. For example, developers should wrap the argument for tf.log() with tf.clip() to avoid the argument turning to zero."
        )
    }

    options = ()

    _variables_with_processing_operation = {}

    def visit_module(self, module: astroid.Module):
        try:
            self._variables_with_processing_operation = TypeInference.infer_variable_full_types(module)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)

    def visit_call(self, call_node: astroid.Call):
        """
        Visit call node to see whether there are rules violations.
        :param call_node:
        :return:
        """
        # if log is call but no mask outside of it, it violate the rule
        try:
            _has_log = False
            _has_mask = False
            if(
                hasattr(call_node.func, "attrname")
                and call_node.func.attrname == "log"
                and hasattr(call_node.func, "expr")
                and hasattr(call_node.func.expr, "name")
                and call_node.func.expr.name in ["tf", "tensorflow"]
            ):
                _has_log = True
            if(
                hasattr(call_node, "args")
                and len(call_node.args) > 0
                and hasattr(call_node.args[0], "func")
                and hasattr(call_node.args[0].func, "attrname")
                and call_node.args[0].func.attrname == "clip_by_value"
            ):
                _has_mask = True
            if(
                hasattr(call_node, "args")
                and len(call_node.args) > 0
                and hasattr(call_node.args[0], "name")
                and call_node.args[0].name in self._variables_with_processing_operation
                and "tf.clip_by_value" in self._variables_with_processing_operation[call_node.args[0].name]
            ):
                _has_mask = True

            if _has_log is True and _has_mask is False:
                self.add_message(msgid="missing-mask-tensorflow", node=call_node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, call_node)



""" In-Place Checker for NumPy which checker In-Place APIs are correctly used. """
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.inplace_helper import inplace_is_true


class InPlaceNumpyChecker(BaseChecker):
    """ In-Place Checker for NumPy which checker In-Place APIs are correctly used. """

    __implements__ = IAstroidChecker

    name = "inplace-numpy"
    priority = -1
    msgs = {
        "W9999": (
            "The operation result has not been assigned to another variable, which might cause losing the result.",
            "inplace-numpy",
            "The result of the operation should be assigned to another variable, or the `out` parameter should be defined."
        )
    }
    options = ()

    def visit_call(self, node: astroid.Call):
        """
        When a Call node is visited, add messages if it violated the defined rules.

        :param node: call node
        """
        try:
            if self._result_is_lost(node):
                self.add_message("inplace-numpy", node=node)
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)

    def _result_is_lost(self, node: astroid.Call) -> bool:
        """
        If the inplace parameter is not set in the API, and the result of API call is
        not assigned to another variable, the result is lost.

        :param node:
        :return: True if the result is lost.
        """
        return (
                hasattr(node, "func")
                and hasattr(node.func, "expr")
                and hasattr(node.func.expr, "name")
                and node.func.expr.name in ["np", "numpy"]  # This might need to be changed to a more robust check
                and not self._function_whitelisted(node)
                and not inplace_is_true(node, "out")
                # If the parent of the Call is an Expression (not an Assignment or a Call or Return),
                # it means the result is lost.
                and isinstance(node.parent, astroid.Expr)
        )

    def _function_whitelisted(self, node) -> bool:
        """
        If the function is in the whitelist, it can be excluded from the check.
        :return:
        """
        WHITELIST = [
            "resize",
            "sort",
            "save",
            "savez",
            "savez_compressed",
            "savetxt",
            "tofile",
            "fill_diagonal",
            "testing",
            "allclose",
            "copyto",
            "set_printoptions",
            "isclose",
            "seterr",
            "scatter_add"
        ]
        return hasattr(node.func, "attrname") and (node.func.attrname in WHITELIST)



"""Checker which checks whether random seed is set in pytorch dataloader"""
from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker
import astroid

from dslinter.utils.exception_handler import ExceptionHandler


class RandomnessControlDataloaderPytorchChecker(BaseChecker):
    """Checker which checks whether random seed is set in pytorch dataloader"""
    __implements__ = IAstroidChecker

    name = "randomness-control-dataloader-pytorch"
    priority = -1
    msgs = {
        "W5512": (
            "The worker_init_fn() and generator is not set in PyTorch DataLoader API",
            "randomness-control-dataloader-pytorch",
            "Use worker_init_fn() and generator in PyTorch DataLoader API to preserve reproducibility"
        )
    }
    options = ()

    _import_dataloader = False

    def visit_importfrom(self, importfrom_node: astroid.ImportFrom):
        """
        Check whether there is DataLoader imported.
        """
        try:
            if(
                hasattr(importfrom_node, "modname")
                and importfrom_node.modname == "torch.utils.data"
                and hasattr(importfrom_node, "names")
            ):
                for name, _ in importfrom_node.names:
                    if name == "DataLoader":
                        self._import_dataloader = True
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, importfrom_node)

    def visit_call(self, node: astroid.Call):
        """
        Check whether there is a rule violation.
        :param node:
        """
        try:
            if self._use_dataloader_from_import(node) or self._use_dataloader_from_torch(node):
                # In dataloader, check if "worker_init_fn" and "generator" is set.
                keywords = []
                if hasattr(node, "keywords"):
                    for k in node.keywords:
                        if hasattr(k, "arg"):
                            keywords.append(k.arg)
                if "worker_init_fn" not in keywords or "generator" not in keywords:
                    self.add_message("randomness-control-dataloader-pytorch", node=node)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, node)

    def _use_dataloader_from_import(self, node):
        # Dataloader has been imported from torch.utils.data
        if(
            self._import_dataloader is True
            and hasattr(node.func, "name")
            and node.func.name == "DataLoader"
        ):
            return True
        return False

    def _use_dataloader_from_torch(self, node):
        # Dataloader has not been imported from torch.utils.data
        full_expr = ""
        if hasattr(node, "func"):
            node = node.func
            while hasattr(node, "expr"):
                if hasattr(node, "attrname"):
                    full_expr = node.attrname + "." + full_expr
                node = node.expr
            if hasattr(node, "name"):
                full_expr = node.name + "." + full_expr
            if full_expr[:-1] == "torch.utils.data.DataLoader":
                return True
        return False



"""Get all parameters of the learning algorithms in pytorch."""
# pylint: disable = line-too-long
# pylint: disable = import-error
from tensorflow.keras.optimizers import Adadelta, Adagrad, Adam, Adamax, Ftrl, Nadam, RMSprop, SGD
from dslinter.scripts.hyperparameters import save_hyperparameter

learning_classes = []
# learning_classes.extend([])  # fit
learning_classes.extend([Adadelta, Adagrad, Adam, Adamax, Ftrl, Nadam, RMSprop, SGD]) #optimizer

if __name__ == "__main__":
    save_hyperparameter(learning_classes, "../resources/hyperparameters_tensorflow_dict.pickle")



"""Class which tests the DataFrameChecker."""
import astroid
import pylint.testutils
import dslinter


class TestDataFrameChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the DataFrameChecker."""

    CHECKER_CLASS = dslinter.plugin.InPlacePandasChecker

    DF_INIT = "import pandas as pd\ndf = pd.DataFrame()\n"

    def test_dataframe_call_assigned(self):
        """Test whether no message is added when a DataFrame operation is assigned."""
        module_tree = astroid.parse(self.DF_INIT + "a = df.abs()")
        assigned_call = module_tree.body[-1].value
        with self.assertNoMessages():
            self.checker.visit_module(module_tree)
            self.checker.visit_call(assigned_call)

    def test_other_call_not_assigned(self):
        """Test whether no message is added when an operation on another type is not assigned."""
        module_tree = astroid.parse(self.DF_INIT + "''.join([])")
        other_call = module_tree.body[-1].value
        with self.assertNoMessages():
            self.checker.visit_module(module_tree)
            self.checker.visit_call(other_call)

    def test_dataframe_call_not_assigned(self):
        """Test whether a message is added when a DataFrame operation is not assigned."""
        module_tree = astroid.parse(self.DF_INIT + "df.abs()")
        unassigned_call = module_tree.body[-1].value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="inplace-pandas", node=unassigned_call),):
            self.checker.visit_module(module_tree)
            self.checker.visit_call(unassigned_call)

    def test_dataframe_call_not_assigned_inplace(self):
        """Test whether no message is added when an inplace DataFrame operation is not assigned."""
        module_tree = astroid.parse(self.DF_INIT + "df.abs(inplace=True)")
        unassigned_call = module_tree.body[-1].value
        with self.assertNoMessages():
            self.checker.visit_module(module_tree)
            self.checker.visit_call(unassigned_call)

    def test_dataframe_call_assigned_whitelisted(self):
        """Test whether no message is added when a DataFrame operation is whitelisted."""
        module_tree = astroid.parse(self.DF_INIT + "df.describe()")
        assigned_call = module_tree.body[-1].value
        with self.assertNoMessages():
            self.checker.visit_module(module_tree)
            self.checker.visit_call(assigned_call)

    def test_dataframe_call_assigned_double(self):
        """Test whether no message is added when two DataFrame operations are assigned."""
        module_tree = astroid.parse(self.DF_INIT + "a = df.abs().abs()")
        assigned_call = module_tree.body[-1].value
        second_call = assigned_call.func.expr
        with self.assertNoMessages():
            self.checker.visit_module(module_tree)
            self.checker.visit_call(assigned_call)
            self.checker.visit_call(second_call)

    def test_dataframe_call_not_assigned_double(self):
        """
        Test whether no message is added when two DataFrame operations are not assigned.

        This is the known false negative.
        """
        module_tree = astroid.parse(self.DF_INIT + "df.abs().abs()")
        unassigned_call = module_tree.body[-1].value
        second_call = unassigned_call.func.expr
        with self.assertNoMessages():
            self.checker.visit_module(module_tree)
            self.checker.visit_call(unassigned_call)
            self.checker.visit_call(second_call)

    def test_dataframe_two_calls_assigned_single_line(self):
        """No message should be added when two DataFrame operations are assigned on one line."""
        module_tree = astroid.parse(self.DF_INIT + "a = df.abs(); b = df.abs()")
        assigned_call_1 = module_tree.body[-2].value
        assigned_call_2 = module_tree.body[-1].value
        with self.assertNoMessages():
            self.checker.visit_module(module_tree)
            self.checker.visit_call(assigned_call_1)
            self.checker.visit_call(assigned_call_2)

    def test_dataframe_two_calls_single_line(self):  # noqa: D205, D400
        """
        Known false negative: The second call on the same line is not checked, because mypy does not
        return two times the same inferred type on the same line.
        """
        module_tree = astroid.parse(self.DF_INIT + "a = df.abs(); df.abs()")
        assigned_call_1 = module_tree.body[-2].value
        assigned_call_2 = module_tree.body[-1].value
        with self.assertNoMessages():
            self.checker.visit_module(module_tree)
            self.checker.visit_call(assigned_call_1)
            self.checker.visit_call(assigned_call_2)

    def test_other_and_dataframe_not_assigned_single_line(self):  # noqa: D205, D400
        """
        Test whether a message is added when a DataFrame operation is not assigned on the same line
        as an operation on another type of object.
        """
        module_tree = astroid.parse(self.DF_INIT + "''.join([]); df.abs()")
        assigned_call_1 = module_tree.body[-2].value
        assigned_call_2 = module_tree.body[-1].value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="inplace-pandas", node=assigned_call_2),):
            self.checker.visit_module(module_tree)
            self.checker.visit_call(assigned_call_1)
            self.checker.visit_call(assigned_call_2)

    def test_dataframe_not_assigned_and_other_single_line(self):  # noqa: D205, D400
        """
        Test whether a message is added when an operation on another type of object is done on the
        same line as an unassigned DataFrame operation.
        """
        module_tree = astroid.parse(self.DF_INIT + "df.abs(); ''.join([])")
        assigned_call_1 = module_tree.body[-2].value
        assigned_call_2 = module_tree.body[-1].value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="inplace-pandas", node=assigned_call_1),):
            self.checker.visit_module(module_tree)
            self.checker.visit_call(assigned_call_1)
            self.checker.visit_call(assigned_call_2)

    def test_dataframe_call_not_assigned_indented_false_negative(self):
        """
        Show the false negative of a violation in an indented setting.

        Due to https://github.com/PyCQA/astroid/issues/784
        """
        module_tree = astroid.parse(self.DF_INIT + "df.abs(\n\ta=0\n)")
        unassigned_call = module_tree.body[-1].value
        with self.assertNoMessages():
            self.checker.visit_module(module_tree)
            self.checker.visit_call(unassigned_call)



"""Checker which checks rules for controlling randomness."""
from typing import List
import traceback
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.resources import Resources


class RandomnessControlScikitLLearnChecker(BaseChecker):
    """Checker which checks rules for controlling randomness."""

    __implements__ = IAstroidChecker

    name = "randomness-control-scikitlearn"
    priority = -1
    msgs = {
        "W5509": (
            "The 'random_state' should be set in estimators or cross-validation splitters.",
            "randomness-control-scikitlearn",
            "For reproducible results across executions, set 'random_state'."
        ),
    }
    options = ()

    SPLITTER_FUNCTIONS: List[str] = [
        "make_classification",
        # "check_cv", # no random_state
        "train_test_split",
    ]

    SPLITTER_CLASSES = [
        # "GroupKFold",
        "GroupShuffleSplit",
        "KFold",
        # "LeaveOneGroupOut",
        # "LeavePGroupsOut",
        # "LeaveOneOut",
        # "LeavePOut",
        # "PredefinedSplit",
        "RepeatedKFold",
        "RepeatedStratifiedKFold",
        "ShuffleSplit",
        "StratifiedKFold",
        "StratifiedShuffleSplit",
        # "TimeSeriesSplit"
    ]

    _HYPERPARAMETER_RESOURCE = "hyperparameters_scikitlearn_dict.pickle"
    _estimators_all = Resources.get_hyperparameters(_HYPERPARAMETER_RESOURCE)

    def visit_call(self, node: astroid.Call):
        """
        When a Call node is visited, check whether it violated the rules in this checker.

        :param node: The node which is visited.
        """

        try:
            if (
                # pylint: disable = R0916
                hasattr(node, "func")
                and hasattr(node, "keywords")
                and hasattr(node.func, "name")
                and (node.func.name in self.SPLITTER_FUNCTIONS
                     or node.func.name in self.SPLITTER_CLASSES)
                     #or node.func.name in self.estimators_all)
            ):
                if node.keywords is not None:
                    _has_random_state_keyword = False
                    for keyword in node.keywords:
                        if keyword.arg == "random_state":
                            _has_random_state_keyword = True
                            if keyword.value.as_string() == "None":
                                self.add_message("randomness-control-scikitlearn", node=node)
                    if _has_random_state_keyword is False:
                        self.add_message("randomness-control-scikitlearn", node=node)

        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, node)
            traceback.print_exc()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dslinter',
 'dslinter.checkers',
 'dslinter.checkers.deprecated',
 'dslinter.resources',
 'dslinter.scripts',
 'dslinter.tests',
 'dslinter.tests.checkers',
 'dslinter.tests.utils',
 'dslinter.utils']

package_data = \
{'': ['*']}

install_requires = \
['astroid==2.9.3',
 'data-science-types==0.2.23',
 'mypy==0.931',
 'pylint==2.12.2',
 'pyspark-stubs==3.0.0.post3',
 'toml>=0.9,<0.10']

setup_kwargs = {
    'name': 'dslinter',
    'version': '2.0.6',
    'description': '`dslinter` is a pylint plugin for linting data science and machine learning code. We plan to support the following Python libraries: TensorFlow, PyTorch, Scikit-Learn, Pandas, NumPy and SciPy.',
    'long_description': '# dslinter\n[![build](https://github.com/SERG-Delft/dslinter/actions/workflows/build.yml/badge.svg)](https://github.com/SERG-Delft/dslinter/actions/workflows/build.yml)\n[![codecov.io](https://codecov.io/github/SERG-Delft/dslinter/coverage.svg?branch=main)](https://codecov.io/github/SERG-Delft/dslinter?branch=main)\n[![PyPI version](https://badge.fury.io/py/dslinter.svg)](https://badge.fury.io/py/dslinter)\n[![License: GPL v3](https://img.shields.io/badge/License-GPL%20v3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)\n[![PyPI - Downloads - Monthly](https://img.shields.io/pypi/dm/dslinter)](https://pypi.org/project/dslinter/) \n[![Code Grade](https://api.codiga.io/project/33224/status/svg)](https://api.codiga.io/project/33224/status/svg)\n\n> Hi! We’re currently researching the code smells in machine learning projects in the industry context and looking for feedback for `dslinter`! It would be a massive help if you could run `dslinter` on your machine learning project in an industry setting and send the text and the json output to dslinter2022@gmail.com . The steps and commands can be found [here](https://github.com/SERG-Delft/dslinter/blob/main/STEPS_TO_FOLLOW.md) and it should take no more than 10 minutes. Feel free to send me an [email](dslinter2022@gmail.com) if you want to go through the process together. The process is anonymous and we will remove any sensitive information before the results are published. Many thanks!\n\n`dslinter` is a PyLint plugin for linting data science and machine learning code. It aims to help developers ensure the machine learning code quality and supports the following Python libraries: TensorFlow, PyTorch, Scikit-Learn, Pandas, NumPy and SciPy. \n\n`dslinter` implements the detection rules for smells identified by [our previous work](https://arxiv.org/pdf/2203.13746.pdf). The smells are collected from papers, grey literature, GitHub commits, and Stack Overflow posts. The smells are also elaborated at a [website](https://hynn01.github.io/ml-smells/) :)\n\n\nhttps://user-images.githubusercontent.com/26082974/166685300-71551c45-6d45-4646-bc34-084239b168bb.mov\n\n\n> The example project in the demo video can be found [here](https://github.com/Hynn01/dslinter-example-projects/tree/main/llexnlp).\n\n## Installation\nTo install from the Python Package Index:\n```\npip install dslinter\n```\n\n## Usage\nTo only use the checkers implemented in this plugin, run:\n\n[For Linux/Mac OS Users]:\n```\npylint \\\n--load-plugins=dslinter \\\n--disable=all \\\n--enable=import,unnecessary-iteration-pandas,unnecessary-iteration-tensorflow,\\\nnan-numpy,chain-indexing-pandas,datatype-pandas,\\\ncolumn-selection-pandas,merge-parameter-pandas,inplace-pandas,\\\ndataframe-conversion-pandas,scaler-missing-scikitlearn,hyperparameters-scikitlearn,\\\nhyperparameters-tensorflow,hyperparameters-pytorch,memory-release-tensorflow,\\\ndeterministic-pytorch,randomness-control-numpy,randomness-control-scikitlearn,\\\nrandomness-control-tensorflow,randomness-control-pytorch,randomness-control-dataloader-pytorch,\\\nmissing-mask-tensorflow,missing-mask-pytorch,tensor-array-tensorflow,\\\nforward-pytorch,gradient-clear-pytorch,data-leakage-scikitlearn,\\\ndependent-threshold-scikitlearn,dependent-threshold-tensorflow,dependent-threshold-pytorch \\\n--output-format=json:report.json,text:report.txt,colorized \\\n--reports=y \\\n<path_to_sources>\n```\n[For Windows Users]:\n```\npylint --load-plugins=dslinter --disable=all --enable=import,unnecessary-iteration-pandas,unnecessary-iteration-tensorflow,nan-numpy,chain-indexing-pandas,datatype-pandas,column-selection-pandas,merge-parameter-pandas,inplace-pandas,dataframe-conversion-pandas,scaler-missing-scikitlearn,hyperparameters-scikitlearn,hyperparameters-tensorflow,hyperparameters-pytorch,memory-release-tensorflow,deterministic-pytorch,randomness-control-numpy,randomness-control-scikitlearn,randomness-control-tensorflow,randomness-control-pytorch,randomness-control-dataloader-pytorch,missing-mask-tensorflow,missing-mask-pytorch,tensor-array-tensorflow,forward-pytorch,gradient-clear-pytorch,data-leakage-scikitlearn,dependent-threshold-scikitlearn,dependent-threshold-tensorflow,dependent-threshold-pytorch --output-format=json:report.json,text:report.txt,colorized --reports=y <path_to_sources>\n```\nOr place a [`.pylintrc` configuration file](https://github.com/Hynn01/dslinter/blob/main/docs/pylint-configuration-examples/pylintrc-with-only-dslinter-settings/.pylintrc) which contains above settings in the folder where you run your command on, and run:\n```\npylint <path_to_sources>\n```\nTo expand a current pylint configuration with the checkers from this plugin, run:\n```\npylint --load-plugins=dslinter <other_options> <path_to_sources>\n```\n\n## How to contribute\nContributions are welcome! If you want to contribute, please see the following steps:\n1. fork the repository and clone the repository you forked.\n```\ngit clone https://github.com/your-github-account/dslinter.git\ngit submodule update --init --recursive\n```\n2. `dslinter` uses `poetry` to manage dependencies, so you will need to install `poetry` first and then install the dependencies. \n```\npip install poetry\npoetry install\n```\n- To install `dslinter` from source for development purposes, install it with:\n```\npoetry build\npip install ./dist/dslinter-version.tar.gz\n```\n3. Assign yourself to the issue you want to solve. If you identify a new issue that needs to be solved, feel free to open a new issue.\n4. Make changes to the repository and run the tests.\nTo run the tests using pytest:\n```\npoetry run pytest .\n```\n5. Make a pull request. The pull request is expected to pass the tests. :)\n\n\n## Implemented Checkers:\n\n- **C5501 - C5506 | import | Import Checker**: Check whether data science modules are imported using the correct naming conventions.\n\n- **R5501 | unnecessary-iteration-pandas | Unnecessary Iteration Checker(Pandas)**: Vectorized solutions are preferred over iterators for DataFrames. If iterations are used while there are vectorized APIs can be used, the rule is violated.\n\n- **W5501 | dataframe-iteration-modification-pandas | Unnecessary Iteration Checker(Pandas)**: A dataframe where is iterated over should not be modified. If the dataframe is modified during iteration, the rule is violated.\n\n- **R5502 | unnecessary-iteration-tensorflow | Unnecessary Iteration Checker(TensorFlow)**: If there is any augment assignment operation in the loop, the rule is violated. Augment assignment in the loop can be replaced with vectorized solution in TensorFlow APIs.\n\n- **E5501 | nan-numpy | Nan Equality Checker(NumPy)**: Values cannot be compared with np.nan, as `np.nan != np.nan`.\n\n- **W5502 | chain-indexing-pandas | Chain Indexing Checker(Pandas)**: Chain indexing is considered bad practice in pandas code and should be avoided. If chain indexing is used on a pandas dataframe, the rule is violated.\n\n- **R5503 | datatype-pandas | Datatype Checker(Pandas)**: Datatype should be set when a dataframe is imported from data to ensure the data formats are imported as expected. If the datatype is not set when importing, the rule is violated.\n\n- **R5504 | column-selection-pandas | Column Selection Checker(Pandas)**: Column should be selected after the dataframe is imported for better elaborating what to be expected in the downstream. \n\n- **R5505 | merge-parameter-pandas | Merge Parameter Checker(Pandas)**: Parameters \'how\', \'on\' and \'validate\' should be set for merge operations to ensure the correct usage of merging.\n\n- **W5503 | inplace-pandas | InPlace Checker(Pandas)**: Operations on DataFrames return new DataFrames, and they should be assigned to a variable. Otherwise the result will be lost, and the rule is violated. Operations from the whitelist and with `in_place` parameter set are excluded.\n\n- **W5504 | dataframe-conversion-pandas | Dataframe Conversion Checker(Pandas)**: For dataframe conversion in pandas code, use .to_numpy() instead of .values. If .values is used in pandas code, the rule is violated.\n\n- **W5505 | scaler-missing-scikitlearn | Scaler Missing Checker(ScikitLearn)**: Check whether the scaler is used before every scaling-sensitive operation in scikit-learn codes. Scaling-sensitive operations includes Principal Component Analysis (PCA), Support Vector Machine (SVM), Stochastic Gradient Descent (SGD), Multi-layer Perceptron classifier and L1 and L2 regularization.\n\n- **R5506 | hyperparameters-scikitlearn | Hyperparameter Checker(ScikitLearn)**: For scikit-learn learning algorithms, some important hyperparameters should be set.\n\n- **R5507 | hyperparameter-tensorflow | Hyperparameter Checker(TensorFlow)**: For neural network learning algorithm, some imporatnt hyperparameters should be set, such as learning rate, batch size, momentum and weight decay.\n\n- **R5508 | hyperparameter-pytorch | Hyperparameter Checker(PyTorch)**: For neural network learning algorithm, some imporatnt hyperparameters should be set, such as learning rate, batch size, momentum and weight decay.\n\n- **W5506 | memory-release-tensorflow | Memory Release Checker(TensorFlow)**: If a neural network is created in the loop, and no memory clear operation is used, the rule is violated.\n\n- **W5507 | deterministic-pytorch | Deterministic Algorithm Usage Checker(PyTorch)**: If use_deterministic algorithm is not used in a pytorch program, the rule is violated.\n\n- **W5508 | randomness-control-numpy | Randomness Control Checker(NumPy)**: The np.seed() should be used to preserve reproducibility in a machine learning program.\n\n- **W5509 | randomness-control-scikitlearn | Randomness Control Checker(ScikitLearn)**: For reproducible results across executions, remove any use of random_state=None in scikit-learn estimators.\n\n- **W5510 | randomness-control-tensorflow | Randomness Control Checker(TensorFlow)**: The tf.random.set_seed() should be used to preserve reproducibility in a Tensorflow program.\n\n- **W5511 | randomness-control-pytorch | Randomness Control Checker(PyTorch)**: The torch.manual_seed() should be used to preserve reproducibility in a Tensorflow program.\n\n- **W5512 | randomness-control-dataloader-pytorch | Randomness Control Checker(PyTorch-Dataloader)**: The worker_init_fn() and generator should be set in dataloader to preserve reproducibility. If they\'re not set, the rule is violated.\n\n- **W5513 | missing-mask-tensorflow | Mask Missing Checker(TensorFlow)**: If log function is used in the code, check whether the argument value is valid. \n\n- **W5514 | missing-mask-pytorch | Mask Missing Checker(PyTorch)**: If log function is used in the code, check whether the argument value is valid. \n\n- **W5515 | tensor-array-tensorflow | Tensor Array Checker(Tensorflow)**: Use tf.TensorArray() for growing array in the loop.\n\n- **W5516 | forward-pytorch | Net Forward Checker(PyTorch)**: It is recommended to use self.net() rather than self.net.forward() in PyTorch code. If self.net.forward() is used in the code, the rule is violated.\n\n- **W5517 | gradient-clear-pytorch | Gradient Clear Checker(PyTorch)**: The loss_fn.backward() and optimizer.step() should be used together with optimizer.zero_grad(). If the ".backward()" is missing in the code, the rule is violated.\n\n- **W5518 | data-leakage-scikitlearn | Data Leakage Checker(ScikitLearn)**: All scikit-learn estimators should be used inside Pipelines, to prevent data leakage between training and test data.\n\n- **W5519 | dependent-threshold-scikitlearn | Dependent Threshold Checker(TensorFlow)**: If threshold-dependent evaluation(e.g., f-score) is used in the code, check whether threshold-indenpendent evaluation(e.g., auc) metrics is also used in the code.\n\n- **W5520 | dependent-threshold-tensorflow | Dependent Threshold Checker(PyTorch)**: If threshold-dependent evaluation(e.g., f-score) is used in the code, check whether threshold-indenpendent evaluation(e.g., auc) metrics is also used in the code.\n\n- **W5521 | dependent-threshold-pytorch | Dependent Threshold Checker(ScikitLearn)**: If threshold-dependent evaluation(e.g., f-score) is used in the code, check whether threshold-indenpendent evaluation(e.g., auc) metrics is also used in the code.\n\n## Team\n\nThe dslinter is developed by Mark Haakman and Haiyin Zhang during our master theses at the Software Engineering Research Group (SERG) at TU Delft and ING\'s AI for FinTech Research Lab, supervised by Luís Cruz and Arie van Deursen.\n\nMaintainer: Haiyin Zhang [hynn0633@gmail.com].\n',
    'author': 'Mark Haakman',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/dslinter/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Class which tests RandomnessControllingScikitlearn."""
import astroid
import pylint.testutils
import dslinter.plugin


class TestRandomnessControlScikitLearnChecker(pylint.testutils.CheckerTestCase):
    """Class which tests RandomnessControllingScikitlearn."""

    CHECKER_CLASS = dslinter.plugin.RandomnessControlScikitLLearnChecker

    def test_with_randomness_controlling(self):
        """If there is randomness controlling in the fucntion, no message is added."""
        script = """
            from sklearn.model_selection import KFold
            rng = 0
            kf = KFold(random_state=rng) #@
        """
        assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    def test_without_randomness_controlling(self):
        """If there is no randomness controlling in the function, a violation message is added."""
        script = """
            from sklearn.model_selection import KFold
            kf = KFold(random_state=None) #@
        """
        assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="randomness-control-scikitlearn", node=call_node)):
            self.checker.visit_call(call_node)

    def test_without_randomness_controlling2(self):
        """If there is no randomness controlling in the function, a ciolation message is added."""
        script = """
            from sklearn.model_selection import KFold
            kf = KFold() #@
        """
        assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="randomness-control-scikitlearn", node=call_node)):
            self.checker.visit_call(call_node)



"""Checker which checks whether values are compared with np.nan."""
import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker
from dslinter.utils.exception_handler import ExceptionHandler


class NanNumpyChecker(BaseChecker):
    """Checker which checks whether values are compared with np.nan."""

    __implements__ = IAstroidChecker

    name = "nan-numpy"
    priority = -1
    msgs = {
        "E5501": (
            "Value compared with np.nan.",
            "nan-numpy",
            "Values cannot be compared with np.nan, as np.nan != np.nan.",
        ),
    }
    options = ()

    def visit_compare(self, node: astroid.Compare):
        """
        When a compare node is visited, check whether a comparison is done with np.nan.

        :param node: Node which is visited.
        """
        try:
            for side in (node.left, node.ops[0][1]):
                if isinstance(side, astroid.Attribute) \
                        and side.attrname == "nan" \
                        and side.expr.name == "np":
                    self.add_message("nan-numpy", node=node)
                    return
        except:  # pylint: disable=bare-except
            ExceptionHandler.handle(self, node)



"""Class which tests the ForwardPytorchChecker."""
import astroid
import pylint.testutils

import dslinter


class TestForwardPytorchChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the ForwardPytorchChecker."""

    CHECKER_CLASS = dslinter.plugin.ForwardPytorchChecker

    def test_use_forward(self):
        """Message will be added if the self.net.forward() is used in the code rather than self.net()."""
        script = """
        import torch.nn as nn
        class Net(nn.Module):
            def __init__(self):
                super().__init__()
                self.conv1 = nn.Conv2d(3, 6, 5)
                self.pool = nn.MaxPool2d(2, 2)
                self.conv2 = nn.Conv2d(6, 16, 5)
                self.fc1 = nn.Linear(16 * 5 * 5, 120)
                self.fc2 = nn.Linear(120, 84)
                self.fc3 = nn.Linear(84, 10)
        
            def forward(self, x):
                x = self.pool.forward(F.relu(self.conv1(x))) #@
                x = self.pool(F.relu(self.conv2(x)))
                x = torch.flatten(x, 1) # flatten all dimensions except batch
                x = F.relu(self.fc1(x))
                x = F.relu(self.fc2(x))
                x = self.fc3(x)
                return x
        """
        call_node = astroid.extract_node(script).value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="forward-pytorch", node=call_node)):
            self.checker.visit_call(call_node)

    def test_not_use_forward(self):
        """No message will be added if self.net() is used in the code."""
        script = """
        import torch.nn as nn
        class Net(nn.Module):
            def __init__(self):
                super().__init__()
                self.conv1 = nn.Conv2d(3, 6, 5)
                self.pool = nn.MaxPool2d(2, 2)
                self.conv2 = nn.Conv2d(6, 16, 5)
                self.fc1 = nn.Linear(16 * 5 * 5, 120)
                self.fc2 = nn.Linear(120, 84)
                self.fc3 = nn.Linear(84, 10)
        
            def forward(self, x):
                x = self.pool(F.relu(self.conv1(x))) #@
                x = self.pool(F.relu(self.conv2(x)))
                x = torch.flatten(x, 1) # flatten all dimensions except batch
                x = F.relu(self.fc1(x))
                x = F.relu(self.fc2(x))
                x = self.fc3(x)
                return x
        """
        call_node = astroid.extract_node(script).value
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    def test_use_self_forward(self):
        """No Message will be added if the self.forward() is used in the code."""
        script = """
        def training_step(self, batch, batch_nb):
            idx = batch['idx']
            loss = self.forward(batch)[0] #@
            return {'loss': loss, 'idx': idx}
        """
        call_node = astroid.extract_node(script).value.value
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    def test_use_super_forward(self):
        """No Message will be added if the super().forward() is used in the code."""
        script = """
        class SpatialDropout(nn.Dropout2d):
            def forward(self, x):
                x = x.unsqueeze(2)    # (N, T, 1, K)
                x = x.permute(0, 3, 2, 1)  # (N, K, 1, T)
                x = super(SpatialDropout, self).forward(x)  # (N, K, 1, T), some features are masked #@
                x = x.permute(0, 3, 2, 1)  # (N, T, 1, K)
                x = x.squeeze(2)  # (N, T, K)
                return x        
        """
        call_node = astroid.extract_node(script).value
        with self.assertNoMessages():
            self.checker.visit_call(call_node)



"""Class which tests DependentThresholdScikitLearnChecker."""
import astroid
import pylint.testutils
import dslinter.plugin


class TestDependentThresholdScikitLearnChecker(pylint.testutils.CheckerTestCase):
    """Class which tests DependentThresholdScikitLearnChecker."""

    CHECKER_CLASS = dslinter.plugin.DependentThresholdScikitLearnChecker

    def test_with_auc_used(self):
        script = """
        import numpy as np
        from sklearn.metrics import auc
        y = np.array([1, 1, 2, 2])
        pred = np.array([0.1, 0.4, 0.35, 0.8])
        fpr, tpr, thresholds = metrics.roc_curve(y, pred, pos_label=2)
        auc(fpr, tpr)
        
        from sklearn.metrics import f1_score
        y_true = [0, 1, 2, 0, 1, 2]
        y_pred = [0, 2, 1, 0, 0, 1]
        f1_score(y_true, y_pred, average='macro')
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_with_only_f1_score_used(self):
        script = """
        from sklearn.metrics import f1_score
        y_true = [0, 1, 2, 0, 1, 2]
        y_pred = [0, 2, 1, 0, 0, 1]
        f1_score(y_true, y_pred, average='macro')
        """
        module = astroid.parse(script)
        f1_score_node = module.body[-1].value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="dependent-threshold-scikitlearn", node=f1_score_node)):
            self.checker.visit_module(module)



"""Class which tests MaskMissingPytorchChecker."""
import astroid
import pylint.testutils

import dslinter.plugin


class TestMaskMissingPytorchChecker(pylint.testutils.CheckerTestCase):
    """Class which tests MaskMissingPytorchChecker."""

    CHECKER_CLASS = dslinter.plugin.MaskMissingPytorchChecker

    def test_missing_mask_1(self):
        script = """
        action_mask = torch.log(action_mask) #@
        """
        module = astroid.parse(script)
        node = astroid.extract_node(script)
        call_node = node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="missing-mask-pytorch", node = call_node)):
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)

    def test_missing_mask_2(self):
        script = """
        action_mask = torch.log(action_mask + 1e-9) #@
        """
        module = astroid.parse(script)
        node = astroid.extract_node(script)
        call_node = node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="missing-mask-pytorch", node = call_node)):
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)

    def test_with_mask_1(self):
        script = """
        action_mask = torch.log(torch.clip(action_mask, FLOAT_MIN, FLOAT_MAX))
        """
        module = astroid.parse(script)
        node = astroid.extract_node(script)
        call_node = node.value
        with self.assertNoMessages():
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)

    def test_with_mask_2(self):
        script = """
        action_mask = torch.log(torch.clamp(action_mask, FLOAT_MIN, FLOAT_MAX))
        """
        module = astroid.parse(script)
        node = astroid.extract_node(script)
        call_node = node.value
        with self.assertNoMessages():
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)

    def test_with_mask_3(self):
        script = """
        action_mask = torch.clip(action_mask, FLOAT_MIN, FLOAT_MAX)
        action_mask = torch.log(action_mask) #@
        """
        module = astroid.parse(script)
        node = astroid.extract_node(script)
        call_node = node.value
        with self.assertNoMessages():
            self.checker.visit_module(module)
            self.checker.visit_call(call_node)



"""Checker which checks whether random seed is set in pytorch"""
from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker
import astroid

from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.randomness_control_helper import check_main_module, has_import


class RandomnessControlPytorchChecker(BaseChecker):
    """Checker which checks whether random seed is set in pytorch"""
    __implements__ = IAstroidChecker

    name = "randomness-control-pytorch"
    priority = -1
    msgs = {
        "W5511": (
            "The torch.manual_seed() is not set in PyTorch program",
            "randomness-control-pytorch",
            "The torch.manual_seed() should be set in PyTorch program for reproducible result"
        )
    }

    options = (
        (
            "no_main_module_check_randomness_control_pytorch",
            {
                "default": False,
                "type": "yn",
                "metavar": "<y_or_n>",
                "help": "Check every module whether torch.manual_seed() is used.",
            },
        ),
    )

    def visit_module(self, module: astroid.Module):
        """
        Check whether there is a rule violation.
        :param module:
        """
        try:
            _import_pytorch = False
            _has_pytorch_manual_seed = False

            # if the user wants to only check main module, but the current file is not main module, just return
            _is_main_module = check_main_module(module)
            if self.config.no_main_module_check_randomness_control_pytorch is False and _is_main_module is False:
                return

            # traverse over the node in the module
            for node in module.body:
                if isinstance(node, astroid.Import):
                    if _import_pytorch is False:
                        _import_pytorch = has_import(node, "torch")

                if isinstance(node, astroid.nodes.Expr) and hasattr(node, "value"):
                    call_node = node.value
                    if _has_pytorch_manual_seed is False:
                        _has_pytorch_manual_seed = self._check_pytorch_manual_seed(call_node)

            # check if the rules are violated
            if(
                _import_pytorch is True
                and _has_pytorch_manual_seed is False
            ):
                self.add_message("randomness-control-pytorch", node=module)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)

    @staticmethod
    def _check_pytorch_manual_seed(call_node: astroid.Call):
        if(
            hasattr(call_node, "func")
            and hasattr(call_node.func, "attrname")
            and call_node.func.attrname == "manual_seed"
        ):
            return True
        return False



"""Some helper functions for inplace checkers"""
import astroid


def inplace_is_true(node: astroid.Call, argument: str) -> bool:
    """
    Evaluate whether the call has an 'out==True' keyword argument.

    :param node: Node to check the arguments from.
    :return: True when the call has an 'inplace==True' keyword argument.
    """
    if node.keywords is None:
        return False

    for keyword in node.keywords:
        if keyword.arg == argument:
            return keyword.value.value
    return False



"""Class which tests ModeTogglingPytorchChecker."""
import astroid
import pylint.testutils

import dslinter.plugin


class TestModeTogglingPytorchChecker(pylint.testutils.CheckerTestCase):
    """Class which tests ModeTogglingPytorchChecker."""

    CHECKER_CLASS = dslinter.plugin.ModeTogglingPytorchChecker

    def test_mode_improper_toggling(self):
        """Message will be added when the training mode is not toggling back in time."""
        script = """
        for epoch in range(2):  # loop over the dataset multiple times 
            running_loss = 0.0
            net.train()
            for i, data in enumerate(trainloader, 0): #@
                # get the inputs; data is a list of [inputs, labels]      
                inputs, labels = data
        
                # zero the parameter gradients
                optimizer.zero_grad()
        
                # forward + backward + optimize
                outputs = net(inputs)
                loss = criterion(outputs, labels)
                loss.backward()
                optimizer.step()
        
                # print statistics
                running_loss += loss.item()
                if i % 2000 == 1999:    # print every 2000 mini-batches
                    print(f'[{epoch + 1}, {i + 1:5d}] loss: {running_loss / 2000:.3f}')
                    running_loss = 0.0
                    # validation
                    net.eval()        
        """
        for_node = astroid.extract_node(script)
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="mode-toggling-pytorch", node=for_node)):
            self.checker.visit_for(for_node)

    def test_mode_proper_toggling(self):
        """No message will be added if the training mode is toggling back in time."""
        script = """
        for epoch in range(2):  # loop over the dataset multiple times 
            running_loss = 0.0
            for i, data in enumerate(trainloader, 0): #@
                # get the inputs; data is a list of [inputs, labels]
                net.train()        
                inputs, labels = data
        
                # zero the parameter gradients
                optimizer.zero_grad()
        
                # forward + backward + optimize
                outputs = net(inputs)
                loss = criterion(outputs, labels)
                loss.backward()
                optimizer.step()
        
                # print statistics
                running_loss += loss.item()
                if i % 2000 == 1999:    # print every 2000 mini-batches
                    print(f'[{epoch + 1}, {i + 1:5d}] loss: {running_loss / 2000:.3f}')
                    running_loss = 0.0
                    # validation
                    net.eval()        
        """
        for_node = astroid.extract_node(script)
        with self.assertNoMessages():
            self.checker.visit_for(for_node)



"""Class which tests DependentThresholdTensorflowChecker."""
import astroid
import pylint.testutils
import dslinter.plugin


class TestDependentThresholdTensorflowChecker(pylint.testutils.CheckerTestCase):
    """Class which tests DependentThresholdTensorflowChecker."""
    CHECKER_CLASS = dslinter.plugin.DependentThresholdTensorflowChecker

    def test_with_auc_used(self):
        script = """
            import tensorflow as tf
            from tf.keras.metrics import AUC
            AUC(
                num_thresholds=200, curve='ROC',
                summation_method='interpolation', name=None, dtype=None,
                thresholds=None, multi_label=False, num_labels=None, label_weights=None,
                from_logits=False
            )
            F1Score(
                # num_classes: tfa.types.FloatTensorLike,
                # average: str = None,
                # threshold: Optional[FloatTensorLike] = None,
                # name: str = 'f1_score',
                # dtype: tfa.types.AcceptableDTypes = None
            )             
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_with_only_f1_score_used(self):
        script = """
            import tensorflow.addons as tfa
            from tensorflow.addons.metrics import F1Score
            F1Score(
                # num_classes: tfa.types.FloatTensorLike,
                # average: str = None,
                # threshold: Optional[FloatTensorLike] = None,
                # name: str = 'f1_score',
                # dtype: tfa.types.AcceptableDTypes = None
            )
        """
        module = astroid.parse(script)
        f1_score_node = module.body[-1].value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="dependent-threshold-tensorflow", node=f1_score_node)):
            self.checker.visit_module(module)



"""Class which tests the HyperparameterScikitLearnChecker."""
import astroid
import pylint.testutils
from pylint.testutils import set_config
import dslinter


class TestHyperparameterScikitLearnChecker(pylint.testutils.CheckerTestCase):
    """Class which tests the HyperparameterScikitLearnChecker."""

    CHECKER_CLASS = dslinter.plugin.HyperparameterScikitLearnChecker

    def test_has_keyword_true(self):
        """Test whether the function returns true when the keyword is present."""
        call_node = astroid.extract_node(
            """
            f(keyword0=0, keyword1=1, keyword2=2) #@
            """
        )
        keywords = call_node.keywords

        assert self.checker.has_keywords(keywords, ["keyword1", "keyword2"])

    def test_has_keyword_false(self):
        """Test whether the function returns false when the keyword is not present."""
        call_node = astroid.extract_node(
            """
            f(keyword0=0) #@
            """
        )
        keywords = call_node.keywords

        assert not self.checker.has_keywords(keywords, ["keyword1"])

    def test_has_keyword_empty(self):
        """Test whether the function returns false when there is no keyword present."""
        call_node = astroid.extract_node(
            """
            f() #@
            """
        )
        keywords = call_node.keywords

        assert not self.checker.has_keywords(keywords, ["keyword1"])

    @set_config(strict_hyperparameters_scikitlearn=True)
    def test_strict_keywords_correct(self):
        """No violation when all keywords are set in strict mode."""
        call_node = astroid.extract_node(
            """
            KMeans(n_clusters=8, init='k-means++', n_init=10, max_iter=300, tol=0.0001, \
            precompute_distances='auto', verbose=0, random_state=None, copy_x=True, n_jobs=None, \
            algorithm='auto') #@
            """
        )
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=True)
    def test_strict_arguments_correct(self):
        """No violation when all arguments are set in strict mode."""
        call_node = astroid.extract_node(
            """
            KMeans(8, 'k-means++', 10, 300, 0.0001, 'auto', 0, None, True, None, 'auto') #@
            """
        )
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=True)
    def test_strict_empty(self):
        """Violation when no keywords or arguments are set in strict mode."""
        call_node = astroid.extract_node(
            """
            KMeans() #@
            """
        )
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="hyperparameters-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=True)
    def test_strict_keywords_missing(self):
        """Violation when not all keywords are set in strict mode."""
        call_node = astroid.extract_node(
            """
            KMeans(n_clusters=8, init='k-means++') #@
            """
        )
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="hyperparameters-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=True)
    def test_strict_arguments_missing(self):
        """Violation when not all arguments are set in strict mode."""
        call_node = astroid.extract_node(
            """
            KMeans(8, 'k-means++') #@
            """
        )
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="hyperparameters-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=False)
    def test_non_strict_keywords_all(self):
        """No violation when all keywords are set in non-strict mode, function from main list."""
        call_node = astroid.extract_node(
            """
            NearestNeighbors(n_neighbors=5, radius=1.0, algorithm='auto', leaf_size=30, \
            metric='minkowski', p=2, metric_params=None, n_jobs=None) #@
            """
        )
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=False)
    def test_non_strict_keywords_required(self):
        """No violation when only required keywords are set in non-strict mode, main function."""
        call_node = astroid.extract_node(
            """
            NearestNeighbors(n_neighbors=5) #@
            """
        )
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=False)
    def test_non_strict_arguments_all(self):
        """No violation when all arguments are set in non-strict mode, function from main list."""
        call_node = astroid.extract_node(
            """
            NearestNeighbors(5, 1.0, 'auto', 30, 'minkowski', 2, None, None) #@
            """
        )
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=False)
    def test_non_strict_empty(self):
        """Violation when no keywords or arguments are set in non-strict mode, main function."""
        call_node = astroid.extract_node(
            """
            NearestNeighbors() #@
            """
        )
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="hyperparameters-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=False)
    def test_non_strict_keywords_missing(self):
        """Violation when not all required keywords are set in non-strict mode, main function."""
        call_node = astroid.extract_node(
            """
            ElasticNet(alpha=1.0) #@
            """
        )
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="hyperparameters-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=False)
    def test_non_strict_arguments_missing(self):
        """Violation when not all required arguments are set in non-strict mode, main function."""
        call_node = astroid.extract_node(
            """
            ElasticNet(1.0) #@
            """
        )
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="hyperparameters-scikitlearn", node=call_node),):
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=False)
    def test_non_strict_non_main_keywords_all(self):
        """No violation when all keywords are set in non-strict mode, no main function."""
        call_node = astroid.extract_node(
            """
            KMeans(n_clusters=8, init='k-means++', n_init=10, max_iter=300, tol=0.0001, \
            precompute_distances='auto', verbose=0, random_state=None, copy_x=True, n_jobs=None, \
            algorithm='auto') #@
            """
        )
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameter_scikitlearn=False)
    def test_non_strict_non_main_keywords_one(self):
        """No violation when one keyword is set in non-strict mode, no main function."""
        call_node = astroid.extract_node(
            """
            KMeans(n_neighbors=5) #@
            """
        )
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=False)
    def test_non_strict_non_main_arguments_all(self):
        """No violation when all arguments are set in non-strict mode, no main function."""
        call_node = astroid.extract_node(
            """
            KMeans(8, 'k-means++', 10, 300, 0.0001, 'auto', 0, None, True, None, 'auto') #@
            """
        )
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    @set_config(strict_hyperparameters_scikitlearn=False)
    def test_non_strict_non_main_arguments_required(self):
        """No violation when one arguments is set in non-strict mode,no main function."""
        call_node = astroid.extract_node(
            """
            KMeans(5) #@
            """
        )
        with self.assertNoMessages():
            self.checker.visit_call(call_node)

    # @set_config(strict_hyperparameters=False)
    # def test_non_strict_non_main_empty(self):
    #     """Violation when no keywords or arguments are set in non-strict mode, no main function."""
    #     call_node = astroid.extract_node(
    #         """
    #         KMeans() #@
    #         """
    #     )
    #     with self.assertAddsMessages(pylint.testutils.Message(msg_id="hyperparameters", node=call_node),):
    #         self.checker.visit_call(call_node)



"""Class which test DatatypePandasChecker."""
import astroid
import pylint.testutils

import dslinter.plugin


class TestDatatypePandasChecker(pylint.testutils.CheckerTestCase):
    """Class which test DatatypePandasChecker."""

    CHECKER_CLASS = dslinter.plugin.DatatypePandasChecker

    def test_datatype_not_set(self):
        """Message should be added if the datatyoe is not set when a dataframe is imported from data."""
        script = """
        df = pd.read_csv('data.csv')  #@   
        """
        assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="datatype-pandas", node=call_node)):
            self.checker.visit_call(call_node)

    def test_datatype_set(self):
        """No message should be added if the datatype is already set."""
        script = """
        df = pd.read_csv('data.csv', dtype={'col1': 'str', 'col2': 'int', 'col3': 'float'})   #@
        """
        assign_node = astroid.extract_node(script)
        call_node = assign_node.value
        with self.assertNoMessages():
            self.checker.visit_call(call_node)
